@@ -28,15 +28,13 @@
     POSSIBILITY OF SUCH DAMAGE.
 '''
 
-import gevent.monkey
-gevent.monkey.patch_all()
-
 import simplejson as json
 import gevent
 import gevent.queue
 import requests
 import requests.exceptions
 from datetime import datetime
+from copy import deepcopy
 
 import logging
 LOG = logging.getLogger('python_bayeux')
@@ -50,6 +48,7 @@ class BayeuxClient(object):
         self.oauth_session = \
             requests.Session() if oauth_session is None else oauth_session
         self.shutdown_called = False
+        self.shutdown_completed = False
 
         # Inbound
         self.message_queue = gevent.queue.Queue()
@@ -66,18 +65,26 @@ class BayeuxClient(object):
         # handshake() has a side effect of initializing self.message_counter
         self.handshake()
 
+        self.connect_timeout = None
+        self.disconnect_complete = False
         self.executing = False
         self.stop_greenlets = False
+        self.waiting_for_resubscribe = False
+        self.go_called = False
+        self.exception = None
 
-        self.outbound_greenlets = [
-            gevent.Greenlet(self._subscribe_greenlet),
-            gevent.Greenlet(self._unsubscribe_greenlet),
-            gevent.Greenlet(self._publish_greenlet)
-        ]
+        self.outbound_greenlets = []
+        for method in (self._subscribe_greenlet,
+                       self._unsubscribe_greenlet,
+                       self._publish_greenlet):
+            new_greenlet = gevent.Greenlet(method)
+            new_greenlet.link_exception(self._exception_callback)
+            self.outbound_greenlets.append(new_greenlet)
 
-        self.inbound_greenlets = [
-            gevent.Greenlet(self._connect_greenlet)
-        ]
+        self.inbound_greenlets = []
+        connect_greenlet = gevent.Greenlet(self._connect_greenlet)
+        connect_greenlet.link_exception(self._exception_callback)
+        self.inbound_greenlets.append(connect_greenlet)
 
         self.greenlets = self.outbound_greenlets + self.inbound_greenlets
 
@@ -109,18 +116,21 @@ class BayeuxClient(object):
         # TODO: handle other 'reconnect' values
         if initial_connect_response_payload['successful']:
             # Convert to seconds
-            self.connect_timeout = \
-                initial_connect_response_payload['advice']['timeout'] / 1000.0
+            if 'advice' in initial_connect_response_payload:
+                advice = initial_connect_response_payload['advice']
+                self.connect_timeout = advice['timeout'] / 1000.0
         # TODO: if not successful, then what?
 
     def disconnect(self):
-        return self._send_message({
+        disconnect_response = self._send_message({
             # MUST
             'channel': '/meta/disconnect',
             'clientId': None,
             # MAY
             'id': None
         })
+        self.disconnect_complete = True
+        return disconnect_response
 
     def connect(self, initial=False):
         connect_request_payload = {
@@ -133,14 +143,12 @@ class BayeuxClient(object):
         }
 
         timeout = None if initial else self.connect_timeout
+
         connect_response = self._send_message(
             connect_request_payload,
             timeout=timeout
         )
 
-        if not isinstance(connect_response, list):
-            raise UnexpectedConnectResponseException(str(connect_response))
-
         return connect_response
 
     def _send_message(self, payload, **kwargs):
@@ -170,10 +178,14 @@ class BayeuxClient(object):
             )
         )
 
+        if len(response.text) == 0:
+            return ''
+
         return response.json()
 
     def _connect_greenlet(self):
         connect_response = None
+
         while not self.stop_greenlets:
             try:
                 connect_response = self.connect()
@@ -184,18 +196,23 @@ class BayeuxClient(object):
                     )
                 )
             else:
+                if not isinstance(connect_response, list):
+                    raise UnexpectedConnectResponseException(
+                        str(connect_response)
+                    )
+
                 messages = []
+                handshake_required = False
                 for element in connect_response:
                     channel = element['channel']
 
                     if channel == '/meta/connect':
                         if not element['successful'] and \
-                           element['error'] == '402::Unknown client':
+                           element['error'] == '403::Unknown client':
 
                             # TODO: support handshake advice interval
                             if element['advice']['reconnect'] == 'handshake':
-                                self.handshake()
-                                return []
+                                handshake_required = True
                     else:
                         # We got a push!
                         messages.append(element)
@@ -203,6 +220,10 @@ class BayeuxClient(object):
                 if len(messages) > 0:
                     self.message_queue.put(messages)
 
+                if handshake_required:
+                    self.handshake()
+                    self._resubscribe()
+
     def _execute_greenlet(self):
         self.executing = True
         channel = None
@@ -236,6 +257,10 @@ class BayeuxClient(object):
                 message_queue_messages,
                 datetime.now()
             ))
+
+            while self.waiting_for_resubscribe:
+                gevent.sleep(0.5)
+
             for message_queue_message in message_queue_messages:
                 channel = message_queue_message['channel']
                 for callback in self.subscription_callbacks[channel]:
@@ -256,10 +281,22 @@ class BayeuxClient(object):
 
         self.subscription_callbacks[channel].append(callback)
 
+    def _resubscribe(self):
+        self.waiting_for_resubscribe = True
 
-    def _subscribe_greenlet(self):
+        current_subscriptions = deepcopy(self.subscription_callbacks)
+        self.subscription_callbacks.clear()
+        for channel, callbacks in current_subscriptions.items():
+            for callback in callbacks:
+                self.subscribe(channel, callback)
+
+        self.waiting_for_resubscribe = False
+
+    def _subscribe_greenlet(self, successive_timeout_threshold=20,
+                            timeout_wait=5):
         channel = None
 
+        successive_timeouts = 0
         while True:
             try:
                 subscription_queue_message = self.subscription_queue.get(
@@ -281,11 +318,25 @@ class BayeuxClient(object):
                 'id': None
             }
 
-            subscribe_responses = self._send_message(subscribe_request_payload)
+            subscribe_responses = []
+            try:
+                subscribe_responses = self._send_message(
+                    subscribe_request_payload
+                )
+            except requests.exceptions.ReadTimeout:
+                successive_timeouts += 1
+
+                if successive_timeouts > successive_timeout_threshold:
+                    raise RepeatedTimeoutException('subscribe')
+
+                gevent.sleep(timeout_wait)
+                self.subscription_queue.put(subscription_queue_message)
+            else:
+                successive_timeouts = 0
 
             for element in subscribe_responses:
                 if not element['successful'] and \
-                   element['error'] == '402::Unknown client':
+                   element['error'] == '403::Unknown client':
                     # Just try again, and eventually connect() will re-try a
                     # handshake
                     self.subscription_queue.put(subscription_queue_message)
@@ -296,7 +347,9 @@ class BayeuxClient(object):
         ))
         self.unsubscription_queue.put(subscription)
 
-    def _unsubscribe_greenlet(self):
+    def _unsubscribe_greenlet(self, successive_timeout_threshold=20,
+                              timeout_wait=5):
+        successive_timeouts = 0
         while True:
             unsubscription = None
             try:
@@ -316,7 +369,18 @@ class BayeuxClient(object):
                 'id': None
             }
 
-            self._send_message(unsubscribe_request_payload)
+            try:
+                self._send_message(unsubscribe_request_payload)
+            except requests.exceptions.ReadTimeout:
+                successive_timeouts += 1
+
+                if successive_timeouts > successive_timeout_threshold:
+                    raise RepeatedTimeoutException('unsubscribe')
+
+                gevent.sleep(timeout_wait)
+                self.unsubscription_queue.put(unsubscription)
+            else:
+                successive_timeouts = 0
 
     def publish(self, channel, payload):
         self.publication_queue.put({
@@ -347,8 +411,9 @@ class BayeuxClient(object):
                 'id': None
             }
 
-            # TODO: at least check for failures
+            # Directly raise exceptions
             publish_response = self._send_message(publish_request_payload)
+
             LOG.info('publish response: {0}'.format(str(publish_response)))
 
     def start(self):
@@ -360,22 +425,28 @@ class BayeuxClient(object):
     # This is how a client can be given its own execute greenlet, but not
     # block the main greenlet
     def go(self):
+        if self.go_called:
+            return
+
+        self.go_called = True
         block_greenlet = gevent.Greenlet(self.block)
+        block_greenlet.link_exception(self._exception_callback)
         self.greenlets.append(block_greenlet)
         block_greenlet.start()
+        # give the execute greenlet a chance to start, so self.executing is
+        # True if we call block() later
+        gevent.sleep(0.1)
 
-    # This is how the main greenlet can be made to block
     def block(self):
         if not self.executing:
             self._execute_greenlet()
         else:
-            # If we have already given this client its own execute greenlet
-            # via go(), then this should do nothing but block
-            while True:
-                if self.stop_greenlets:
-                    break
-                else:
-                    gevent.sleep(1)
+            # block the main greenlet
+            while self.exception is None and any(self.greenlets):
+                gevent.sleep(1)
+
+        if self.exception is not None:
+            raise self.exception
 
     def shutdown(self):
         if not self.shutdown_called:
@@ -387,16 +458,54 @@ class BayeuxClient(object):
 
             # disconnect() runs in the main greenlet, so we want to give the
             # others a chance to finish
-            gevent.joinall(
-                [greenlet for greenlet in self.greenlets if greenlet]
-            )
+            # If we have been called by a callback (that is, the client wants
+            # to shut down itself), then we don't want to wait for the execute
+            # greenlet to stop, because we'll deadlock.
+            # This also means that subclasses should not change self.greenlets
+            relevant_greenlets = \
+                self.greenlets[:-1] \
+                if gevent.getcurrent() == self.greenlets[-1] \
+                else self.greenlets
+
+            gevent.joinall(relevant_greenlets)
             self.disconnect()
+            self.shutdown_completed = True
+
+    def _exception_callback(self, failed_greenlet):
+        LOG.info(
+            'client id {0} has an unhandled exception '
+            'in greenlet {1}: {2}'.format(
+                self.client_id,
+                failed_greenlet.name,
+                str(failed_greenlet.exception)
+            )
+        )
+
+        self.exception = failed_greenlet.exception
+
+        try:
+            self.shutdown()
+        except Exception:
+            # We prefer to shut down, but if something doesn't work (like the
+            # org is sending back a total requests limit error), then don't
+            # worry
+            pass
 
     def __enter__(self):
         return self
 
     def __exit__(self, exception_type, exception_value, traceback):
         self.shutdown()
+        while not self.shutdown_completed:
+            gevent.sleep(0.5)
+
+
+class RepeatedTimeoutException(Exception):
+    def __init__(self, greenlet_name):
+        self.greenlet_name = greenlet_name
+        super(RepeatedTimeoutException, self).__init__(
+            'Too many timeouts'
+        )
 
 
 class UnexpectedConnectResponseException(Exception):
